;; credits.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(define-module (td credits)
  :use-module (td hooks)
  :use-module (td gfx)
  :use-module (td picture)
  :use-module (chickadee render font)
  :use-module (chickadee math)
  :use-module (chickadee math vector)
  :use-module (oop goops)
  :export (credits
           credits-text
           default-credits
           deposit-credits!
           has-credits?
           set-credits!
           try-purchase!
           withdrawl-credits!))

(define credits 0)

(define (set-credits! amt)
  (set! credits amt))

(add-hook! new-game-hook
           (lambda () (set-credits! default-credits)))

(define (credits-text)
  (make <text>
    #:font font-gns
    #:string (string-append "Credits: "
                            (number->string credits))))

(define default-credits 700)

(define (has-credits? amt)
  (<= amt credits))

(define (withdrawl-credits! amt)
  (set-credits! (- credits amt)))

(define (deposit-credits! amt)
  (set-credits! (+ credits amt)))

(define (try-purchase! amt)
  (if (has-credits? amt)
      (withdrawl-credits! amt)
      (throw 'insufficient-credits)))
