;; -*- geiser-scheme-implementation: guile -*-

;; gfx.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(define-module (td gfx)
  #:use-module (td hooks)
  #:use-module (td picture)
  #:use-module (aa-tree)
  #:use-module (chickadee)
  #:use-module (chickadee math vector)
  #:use-module (chickadee render font)
  #:use-module (chickadee render texture)
  #:use-module (oop goops)
  #:export (font-gns
            frame-rate-text
            get-sprite))

(define font-gns #nil)
(define frame-time-before 30)
(define frame-time-after 30)
(define frame-rate 30)
(define frame-rate-prev 30)
(define sprites)

(define (load-atlases)
  (list-to-aa-tree #nil
                   (list #:t40
                         (split-texture (load-image "images/early/40.png")
                                        40 40)
                         #:t64
                         (split-texture (load-image "images/cmh/towers64.png")
                                        64 64)
                         #:t64b
                         (split-texture (load-image "images/cmh/towers64b.png")
                                        64 64)
                         #:s64
                         (split-texture (load-image "images/cmh/soldiers64.png")
                                        64 64)
                         #:v64
                         (split-texture (load-image "images/cmh/vehicles64.png")
                                        64 64)
                         #:v64b
                         (split-texture (load-image "images/cmh/vehicles64b.png")
                                        64 64)
                         #:v64c
                         (split-texture (load-image "images/cmh/vehicles64c.png")
                                        64 64))))

(define (load-sprites)
  (let ((atlases (load-atlases)))
    (list-to-aa-tree #nil
                     (list #:basic-turret
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 0)
                             #:origin (vec2 32 32))
                           #:basic-turret-fire-0
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 1)
                             #:origin (vec2 32 32))
                           #:basic-turret-fire-1
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 2)
                             #:origin (vec2 32 32))
                           #:basic-turret-fire-2
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 3)
                             #:origin (vec2 32 32))
                           #:basic-turret-ghost
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64) 10)
                             #:origin (vec2 32 32))
                           #:basic-turret-l2
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 8)
                             #:origin (vec2 32 32))
                           #:basic-turret-l2-fire-0
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 9)
                             #:origin (vec2 32 32))
                           #:basic-turret-l2-fire-1
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 10)
                             #:origin (vec2 32 32))
                           #:basic-turret-l2-fire-2
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 11)
                             #:origin (vec2 32 32))
                           #:basic-turret-l2-fire-3
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 12)
                             #:origin (vec2 32 32))
                           #:basic-turret-l2-fire-4
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 13)
                             #:origin (vec2 32 32))
                           #:basic-turret-l3
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 16)
                             #:origin (vec2 32 32))
                           #:basic-turret-l3-fire-0
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 17)
                             #:origin (vec2 32 32))
                           #:basic-turret-l3-fire-1
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 18)
                             #:origin (vec2 32 32))
                           #:basic-turret-l3-fire-2
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 19)
                             #:origin (vec2 32 32))
                           #:basic-turret-l3-fire-3
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 20)
                             #:origin (vec2 32 32))
                           #:basic-turret-l3-fire-4
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 21)
                             #:origin (vec2 32 32))
                           #:basic-turret-l3-fire-5
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 22)
                             #:origin (vec2 32 32))
                           #:basic-turret-l3-fire-6
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 23)
                             #:origin (vec2 32 32))
                           #:basic-turret-l3-fire-7
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 15)
                             #:origin (vec2 32 32))
                           #:defeat
                           (make <sprite>
                             #:texture (load-image "images/cmh/defeat.png")
                             #:origin (vec2 200 200))
                           #:delay-turret
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 24)
                             #:origin (vec2 32 32))
                           #:delay-turret-fire-0
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 25)
                             #:origin (vec2 32 32))
                           #:delay-turret-fire-1
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 26)
                             #:origin (vec2 32 32))
                           #:delay-turret-fire-2
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 27)
                             #:origin (vec2 32 32))
                           #:delay-turret-ghost
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 28)
                             #:origin (vec2 32 32))
                           #:delay-turret-l2
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 32)
                             #:origin (vec2 32 32))
                           #:delay-turret-l2-fire-0
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 33)
                             #:origin (vec2 32 32))
                           #:delay-turret-l2-fire-1
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 34)
                             #:origin (vec2 32 32))
                           #:delay-turret-l3
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 40)
                             #:origin (vec2 32 32))
                           #:delay-turret-l3-fire-0
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64b) 41)
                             #:origin (vec2 32 32))
                           #:destroyer-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 12)
                             #:origin (vec2 32 32))
                           #:destroyer-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 15)
                             #:origin (vec2 32 32))
                           #:destroyer-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 13)
                             #:origin (vec2 32 32))
                           #:destroyer-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 14)
                             #:origin (vec2 32 32))
                           #:field0
                           (make <sprite>
                             #:texture (load-image "maps/rt-based/map0.png")
                             #:origin (vec2 512 384))
                           #:helo-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64b) 4)
                             #:origin (vec2 32 32))
                           #:helo-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64b) 7)
                             #:origin (vec2 32 32))
                           #:helo-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64b) 5)
                             #:origin (vec2 32 32))
                           #:helo-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64b) 6)
                             #:origin (vec2 32 32))
                           #:ptboat-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 8)
                             #:origin (vec2 32 32))
                           #:ptboat-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 11)
                             #:origin (vec2 32 32))
                           #:ptboat-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 9)
                             #:origin (vec2 32 32))
                           #:ptboat-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 10)
                             #:origin (vec2 32 32))
                           #:resource-tower
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t40) 3)
                             #:origin (vec2 20 20))
                           #:resource-tower-ghost
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64) 9)
                             #:origin (vec2 32 32))
                           #:soldier-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 0)
                             #:origin (vec2 32 32))
                           #:soldier-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 3)
                             #:origin (vec2 32 32))
                           #:soldier-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 1)
                             #:origin (vec2 32 32))
                           #:soldier-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 2)
                             #:origin (vec2 32 32))
                           #:soldier-l2-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 8)
                             #:origin (vec2 32 32))
                           #:soldier-l2-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 11)
                             #:origin (vec2 32 32))
                           #:soldier-l2-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 9)
                             #:origin (vec2 32 32))
                           #:soldier-l2-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 10)
                             #:origin (vec2 32 32))
                           #:soldier-l3-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 16)
                             #:origin (vec2 32 32))
                           #:soldier-l3-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 19)
                             #:origin (vec2 32 32))
                           #:soldier-l3-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 17)
                             #:origin (vec2 32 32))
                           #:soldier-l3-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:s64) 18)
                             #:origin (vec2 32 32))
                           #:resource-tower
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t40) 3)
                             #:origin (vec2 20 20))
                           #:resource-tower-l2
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64) 7)
                             #:origin (vec2 32 32))
                           #:resource-tower-l3
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:t64) 8)
                             #:origin (vec2 32 32))
                           #:tank-blue-t0-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 2)
                             #:origin (vec2 32 32))
                           #:tank-blue-t0-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 1)
                             #:origin (vec2 32 32))
                           #:tank-blue-t0-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 3)
                             #:origin (vec2 32 32))
                           #:tank-blue-t0-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 0)
                             #:origin (vec2 32 32))
                           #:tank-blue-t1-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 6)
                             #:origin (vec2 32 32))
                           #:tank-blue-t1-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 5)
                             #:origin (vec2 32 32))
                           #:tank-blue-t1-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 7)
                             #:origin (vec2 32 32))
                           #:tank-blue-t1-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 4)
                             #:origin (vec2 32 32))
                           #:tank-blue-t2-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 10)
                             #:origin (vec2 32 32))
                           #:tank-blue-t2-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 9)
                             #:origin (vec2 32 32))
                           #:tank-blue-t2-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 11)
                             #:origin (vec2 32 32))
                           #:tank-blue-t2-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 8)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t0-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 18)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t0-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 17)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t0-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 19)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t0-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 16)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t1-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 22)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t1-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 21)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t1-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 23)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t1-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 20)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t2-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 26)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t2-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 25)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t2-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 27)
                             #:origin (vec2 32 32))
                           #:tank-yellow-t2-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 24)
                             #:origin (vec2 32 32))
                           #:tank-red-t0-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 34)
                             #:origin (vec2 32 32))
                           #:tank-red-t0-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 33)
                             #:origin (vec2 32 32))
                           #:tank-red-t0-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 35)
                             #:origin (vec2 32 32))
                           #:tank-red-t0-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 32)
                             #:origin (vec2 32 32))
                           #:tank-red-t1-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 38)
                             #:origin (vec2 32 32))
                           #:tank-red-t1-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 37)
                             #:origin (vec2 32 32))
                           #:tank-red-t1-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 39)
                             #:origin (vec2 32 32))
                           #:tank-red-t1-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 36)
                             #:origin (vec2 32 32))
                           #:tank-red-t2-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 42)
                             #:origin (vec2 32 32))
                           #:tank-red-t2-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 41)
                             #:origin (vec2 32 32))
                           #:tank-red-t2-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 43)
                             #:origin (vec2 32 32))
                           #:tank-red-t2-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64c) 40)
                             #:origin (vec2 32 32))
                           #:trike-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 0)
                             #:origin (vec2 32 32))
                           #:tracked-soldier-east
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64b) 0)
                             #:origin (vec2 32 32))
                           #:tracked-soldier-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64b) 3)
                             #:origin (vec2 32 32))
                           #:tracked-soldier-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64b) 1)
                             #:origin (vec2 32 32))
                           #:tracked-soldier-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64b) 2)
                             #:origin (vec2 32 32))
                           #:trike-north
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 3)
                             #:origin (vec2 32 32))
                           #:trike-south
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 1)
                             #:origin (vec2 32 32))
                           #:trike-west
                           (make <sprite>
                             #:texture (texture-atlas-ref
                                        (aa-search atlases #:v64) 2)
                             #:origin (vec2 32 32))
                           #:victory
                           (make <sprite>
                             #:texture (load-image "images/cmh/victory.png")
                             #:origin (vec2 200 200))
                           ))))

(define (load-sprites!)
  (set! sprites (load-sprites)))

(define (get-sprite key)
  (aa-search sprites key))

(add-hook! load-hook (lambda () (load-sprites!)))

(add-hook! load-hook
           (lambda ()
             (set! font-gns
               (load-font "font/gn/good_neighbors_starling.xml"))
             (run-hook fonts-loaded-hook)))
    
(add-hook! before-draw-hook
           (lambda ()
             (set! frame-time-before frame-time-after)
             (set! frame-time-after (tms:clock (times)))
             (set! frame-rate-prev frame-rate)
             (set! frame-rate
               (catch 'numerical-overflow
                 (lambda ()
                   (+ (* frame-rate-prev 0.95)
                      (* (/ 1000000000.0
                            (- frame-time-after frame-time-before))
                         0.05)))
                 (lambda _ 99)))))

(define (frame-rate-text)
  (make <text>
    #:font font-gns
    #:string (string-append "FPS: "
                            (number->string
                             (inexact->exact (round frame-rate))))))
