;; -*- geiser-scheme-implementation: guile -*-

;; picture.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; The goal of this module is to provide an overarching interface to
;; drawable objects, similar to the Haskell Gloss Picture interface.

(define-module (td picture)
  #:use-module (td fset)
  #:use-module (chickadee)
  #:use-module (chickadee math vector)
  #:use-module (chickadee render color)
  #:use-module (chickadee render font)
  #:use-module (chickadee render shapes)
  #:use-module (chickadee render sprite)
  #:use-module (chickadee render texture)
  #:use-module (oop goops)
  #:export (draw
            transform
            <line>
            <sprite>
            <text>))

(define-class <sprite> ()
  (texture #:accessor .texture #:init-keyword #:texture)
  (origin #:accessor .origin #:init-keyword #:origin)
  (shift #:accessor shift #:init-keyword #:shift)
  ;; slots for scale, rotation, etc.
  )

(define-method (transform (s <sprite>) v)
  (if (slot-bound? s 'shift)
      (let ((new-shift (vec2-add! (vec2-copy (shift s))
                                  v)))
        (slot-fset s 'shift new-shift))
      (slot-fset s 'shift v)))

(define-method (draw (s <sprite>) position)
  (if (slot-bound? s 'shift)
      (let ((new-position (vec2-copy position)))
        (begin
          (vec2-add! new-position (shift s))
          (draw-sprite (.texture s) new-position #:origin (.origin s))))
      (draw-sprite (.texture s) position #:origin (.origin s))))

(define-class <line> ()
  (start #:accessor .start #:init-keyword #:start)
  (end #:accessor .end #:init-keyword #:end)
  (thickness #:accessor .thickness #:init-keyword #:thickness #:init-value 1)
  (color #:accessor .color #:init-keyword #:color #:init-value white))

(define reg-dl-start (vec2 0 0))
(define reg-dl-end (vec2 0 0))

(define-method (draw (l <line>) position)
  (vec2-copy! position reg-dl-start)
  (vec2-copy! position reg-dl-end)
  (vec2-add! reg-dl-start (.start l))
  (vec2-add! reg-dl-end (.end l))
  (draw-line reg-dl-start
             reg-dl-end
             #:thickness (.thickness l)
             #:color (.color l)))

(define-class <text> ()
  (string #:accessor .string #:init-keyword #:string)
  (font #:accessor .font #:init-keyword #:font))

(define-method (draw (t <text>) position)
  (draw-text (.font t) (.string t) position))
